use std::f32::consts::PI;

use bevy::core::{FixedTimestep, Time, Timer};
use bevy::input::system::exit_on_esc_system;
use bevy::input::Input;
use bevy::math::Vec3;
use bevy::prelude::*;
use bevy::render::primitives::Frustum;
use bevy::sprite::{Sprite, SpriteBundle};
use bevy::text::Text;
use bevy::window::Windows;
use bevy::DefaultPlugins;
use rand::Rng;

fn main() {
    App::new()
        .add_plugins(DefaultPlugins)
        .insert_resource(ArnoldStats::new())
        .add_startup_system(setup)
        .add_system_set(
            SystemSet::new()
                .with_run_criteria(FixedTimestep::step(1.0 / 60.0))
                .with_system(projectile_collide_system)
                .with_system(arnold_movement_system)
                .with_system(arnold_attack_system)
                .with_system(linear_movement_system)
                .with_system(expiration_date_system)
                .with_system(follow_arnold_system)
                .with_system(enemy_spawner_system)
                .with_system(update_scoreboard_system),
        )
        .add_system(exit_on_esc_system)
        .run();
}

fn setup(mut commands: Commands, assets: Res<AssetServer>) {
    let arnold_sh = assets.load("arno.png");

    // cameras
    commands.spawn_bundle(OrthographicCameraBundle::new_2d());
    commands.spawn_bundle(UiCameraBundle::default());
    // background

    for x in -20..20 {
        for y in -20..20 {
            commands.spawn_bundle(SpriteBundle {
                texture: assets.load("grass.png"),
                transform: Transform {
                    translation: Vec3::new(x as f32 * 128.0, y as f32 * 128.0, 0.0),
                    ..Default::default()
                },
                ..Default::default()
            });
        }
    }
    // arnold himself
    commands
        .spawn_bundle(SpriteBundle {
            texture: arnold_sh,
            ..Default::default()
        })
        .insert(Arnold::new())
        .insert(Transform {
            translation: Vec3::new(0.0, 0.0, 10.0),
            ..Default::default()
        });
    // enemy spawner
    commands
        .spawn()
        .insert(EnemySpawner::new(Timer::from_seconds(1.5, true)));
    // scoreboard
    commands
        .spawn_bundle(TextBundle {
            text: Text {
                sections: vec![
                    TextSection {
                        style: TextStyle {
                            font: assets.load("fonts/PressStart2P-Regular.ttf"),
                            font_size: 15.0,
                            color: Color::WHITE,
                        },
                        value: "kills: ".to_string(),
                    },
                    TextSection {
                        style: TextStyle {
                            font: assets.load("fonts/PressStart2P-Regular.ttf"),
                            font_size: 15.0,
                            color: Color::RED,
                        },
                        value: "0".to_string(),
                    },
                ],
                ..Default::default()
            },
            style: Style {
                position_type: PositionType::Absolute,
                position: Rect {
                    top: Val::Px(5.0),
                    left: Val::Px(5.0),
                    ..Default::default()
                },
                ..Default::default()
            },
            ..Default::default()
        })
        .insert(ScoreBoard);
}

// Unit struct for tracking the ScoreBoard Text UI
#[derive(Component)]
struct ScoreBoard;

fn update_scoreboard_system(
    stats: Res<ArnoldStats>,
    mut query: Query<&mut Text, With<ScoreBoard>>,
) {
    query.single_mut().sections[1].value = format!("{}", stats.kills);
}

struct ArnoldStats {
    damage_dealt: i32,
    kills: i32,
}

impl ArnoldStats {
    fn new() -> Self {
        Self {
            damage_dealt: 0,
            kills: 0,
        }
    }
}

#[derive(Component)]
struct Projectile {
    damage: i32,
}

impl Projectile {
    fn new(damage: i32) -> Self {
        Self { damage }
    }
}

#[derive(Component)]
struct Hitpoints {
    hp: i32,
}

impl Hitpoints {
    fn new(hp: i32) -> Self {
        Self { hp }
    }
}

fn projectile_collide_system(
    mut commands: Commands,
    mut stats: ResMut<ArnoldStats>,
    assets: Res<AssetServer>,
    projectile_query: Query<(&Projectile, &Transform)>,
    mut target_query: Query<(Entity, &mut Hitpoints, &Transform)>,
) {
    let mut rng = rand::thread_rng();

    for (entity, mut hp, target_transform) in target_query.iter_mut() {
        for (projectile, projectile_transform) in projectile_query.iter() {
            if target_transform
                .translation
                .distance(projectile_transform.translation)
                < 32.0
            {
                let damage_variable = projectile.damage / 5;
                let actual_damage = projectile.damage - damage_variable / 2 + rng.gen_range(0..damage_variable + 1);
                hp.hp -= actual_damage;
                stats.damage_dealt += actual_damage;

                let font = assets.load("fonts/PressStart2P-Regular.ttf");
                let text_style = TextStyle {
                    font,
                    font_size: 10.0,
                    color: Color::YELLOW,
                };
                let text_alignment = TextAlignment {
                    vertical: VerticalAlign::Center,
                    horizontal: HorizontalAlign::Center,
                };
                commands
                    .spawn_bundle(Text2dBundle {
                        text: Text::with_section(
                            &format!("{}", actual_damage),
                            text_style,
                            text_alignment,
                        ),
                        transform: Transform {
                            translation: target_transform.translation + Vec3::new(0.0, 10.0, 10.0),
                            ..Default::default()
                        },
                        ..Default::default()
                    })
                    .insert(ExpirationDate::new(3.0))
                    .insert(LinearMovement::new(Vec3::new(0.0, 0.5, 0.0)));
            }
        }

        if hp.hp <= 0 {
            stats.kills += 1;
            commands.entity(entity).despawn();
        }
    }
}

#[derive(Clone, Copy, PartialEq)]
enum Facing {
    Left,
    Right,
}

#[derive(Component)]
struct Arnold {
    facing: Facing,
    attack_timer: Timer,
}

impl Arnold {
    pub fn new() -> Self {
        Self {
            facing: Facing::Left,
            attack_timer: Timer::from_seconds(1.0, true),
        }
    }
}

#[derive(Component)]
struct Knife {}

impl Knife {
    fn new() -> Self {
        Self {}
    }
}

#[derive(Component)]
struct LinearMovement {
    direction: Vec3,
}

impl LinearMovement {
    fn new(direction: Vec3) -> Self {
        Self { direction }
    }
}

fn linear_movement_system(mut query: Query<(&LinearMovement, &mut Transform)>) {
    for (lin_movement, mut transform) in query.iter_mut() {
        transform.translation += lin_movement.direction;
    }
}

#[derive(Component)]
struct ExpirationDate {
    countdown: Timer,
}

impl ExpirationDate {
    fn new(duration: f32) -> Self {
        Self {
            countdown: Timer::from_seconds(duration, false),
        }
    }
}

fn expiration_date_system(
    mut commands: Commands,
    time: Res<Time>,
    mut query: Query<(Entity, &mut ExpirationDate)>,
) {
    for (entity, mut exp_date) in query.iter_mut() {
        if exp_date.countdown.tick(time.delta()).just_finished() {
            commands.entity(entity).despawn();
        }
    }
}

fn arnold_movement_system(
    keyboard_input: Res<Input<KeyCode>>,
    mut query: Query<(&mut Arnold, &mut Sprite, &mut Transform), Without<Camera>>,
    // Hack to get exactly the right camera (the one for the given 2d scene, not the UI one)
    mut cam_query: Query<(&Camera, &mut Transform), With<Frustum>>,
) {
    let (_cam, mut cam_transform) = cam_query.single_mut();
    let (mut arnold, mut sprite, mut transform) = query.single_mut();

    let mut velocity = Vec3::new(0.0, 0.0, 0.0);
    if keyboard_input.pressed(KeyCode::A) {
        velocity.x -= 1.0;
    }
    if keyboard_input.pressed(KeyCode::D) {
        velocity.x += 1.0;
    }
    if keyboard_input.pressed(KeyCode::S) {
        velocity.y -= 1.0;
    }
    if keyboard_input.pressed(KeyCode::W) {
        velocity.y += 1.0;
    }
    velocity *= 1.5;

    if keyboard_input.just_pressed(KeyCode::A) {
        sprite.flip_x = false;
        arnold.facing = Facing::Left;
    }
    if keyboard_input.just_pressed(KeyCode::D) {
        sprite.flip_x = true;
        arnold.facing = Facing::Right;
    }

    transform.translation += velocity;
    cam_transform.translation += velocity;
}

fn arnold_attack_system(
    assets: Res<AssetServer>,
    mut commands: Commands,
    time: Res<Time>,
    mut query: Query<(&mut Arnold, &Transform)>,
) {
    let (mut arnold, arnold_transform) = query.single_mut();

    if arnold.attack_timer.tick(time.delta()).just_finished() {
        commands.spawn_bundle(KnifeBundle::new(
            assets.load("knife.png"),
            arnold_transform.translation + Vec3::new(0.0, 0.0, 10.0),
            arnold.facing,
            4.0,
            5.0,
            10,
        ));
    }
}

#[derive(Component)]
struct EnemySpawner {
    timer: Timer,
}

impl EnemySpawner {
    fn new(timer: Timer) -> Self {
        Self { timer }
    }
}

fn enemy_spawner_system(
    mut commands: Commands,
    assets: Res<AssetServer>,
    windows: Res<Windows>,
    time: Res<Time>,
    mut query: Query<&mut EnemySpawner>,
    query_player: Query<&Transform, With<Arnold>>,
) {
    let mut enemy_spawner = query.single_mut();

    if !enemy_spawner.timer.tick(time.delta()).just_finished() {
        return;
    }

    let window = windows.get_primary().unwrap();
    let spawn_distance = f32::max(window.width(), window.height()) / 2.0;

    let mut rng = rand::thread_rng();
    let phi = rng.gen_range(0.0..2.0 * PI);

    commands
        .spawn_bundle(SpriteBundle {
            texture: assets.load("monster.png"),
            transform: Transform {
                translation: query_player.single().translation
                    + Vec3::new(phi.cos() * spawn_distance, phi.sin() * spawn_distance, 0.0),
                ..Default::default()
            },
            ..Default::default()
        })
        .insert(FollowArnold::new(0.8))
        .insert(Hitpoints::new(8));
}

#[derive(Bundle)]
struct KnifeBundle {
    _knife: Knife,
    #[bundle]
    sprite: SpriteBundle,
    lin_movement: LinearMovement,
    exp_date: ExpirationDate,
    projectile: Projectile,
}

impl KnifeBundle {
    fn new(
        texture: Handle<Image>,
        translation: Vec3,
        facing: Facing,
        speed: f32,
        lifetime: f32,
        damage: i32,
    ) -> Self {
        let velocity = match facing {
            Facing::Left => Vec3::new(-1.0, 0.0, 0.0),
            Facing::Right => Vec3::new(1.0, 0.0, 0.0),
        } * speed;
        let flip_x = velocity.x < 0.0;

        KnifeBundle {
            _knife: Knife::new(),
            sprite: SpriteBundle {
                texture,
                sprite: Sprite {
                    flip_x,
                    ..Default::default()
                },
                transform: Transform {
                    translation,
                    ..Default::default()
                },
                ..Default::default()
            },
            lin_movement: LinearMovement::new(velocity),
            exp_date: ExpirationDate::new(lifetime),
            projectile: Projectile::new(damage),
        }
    }
}

#[derive(Component)]
struct FollowArnold {
    speed: f32,
}

impl FollowArnold {
    fn new(speed: f32) -> Self {
        Self { speed }
    }
}

fn follow_arnold_system(
    mut follow_query: Query<(&FollowArnold, &mut Transform), Without<Arnold>>,
    arnold_query: Query<(&Arnold, &Transform)>,
) {
    let (_, arnold_transform) = arnold_query.single();

    for (follow, mut follow_transform) in follow_query.iter_mut() {
        let velocity = (arnold_transform.translation - follow_transform.translation).normalize()
            * follow.speed;
        follow_transform.translation += velocity;
    }
}
